try:
    from redis import StrictRedis
except ImportError:
    StrictRedis = None

from thumbnails import compat, conf
from thumbnails.models import Source, ThumbnailMeta
from thumbnails.utils import import_attribute


def get_backend():
    if not conf.METADATA.get('BACKEND'):
        raise ValueError('BACKEND for STORAGE must be defined')
    metadata = import_attribute(conf.METADATA['BACKEND'])
    return metadata()


class ImageMeta:

    def __init__(self, source_name, name, size):
        self.source_name = source_name
        self.name = compat.as_text(name)
        self.size = compat.as_text(size)

    def __eq__(self, other):
        try:
            return self.__dict__ == other.__dict__
        except AttributeError:
            return False


class BaseBackend:

    def add_source(name):
        raise NotImplementedError

    def delete_source(name):
        raise NotImplementedError

    def get_thumbnails(name):
        raise NotImplementedError

    def get_thumbnail(name, size):
        raise NotImplementedError

    def add_thumbnail(name, size, filename):
        raise NotImplementedError

    def delete_thumbnail(name, size):
        raise NotImplementedError

    def flush_thumbnails(name):
        raise NotImplementedError


class DatabaseBackend(BaseBackend):

    def add_source(self, name):
        return Source.objects.create(name=name)

    def get_source(self, name):
        return Source.objects.get(name=name)

    def delete_source(self, name):
        return Source.objects.filter(name=name).delete()

    def get_thumbnails(self, name):
        metas = ThumbnailMeta.objects.filter(source__name=name)
        return [ImageMeta(name, meta.name, meta.size) for meta in metas]

    def get_thumbnail(self, source_name, size):
        try:
            meta = ThumbnailMeta.objects.get(source__name=source_name, size=size)
            return ImageMeta(source_name, meta.name, meta.size)
        except ThumbnailMeta.DoesNotExist:
            return None

    def add_thumbnail(self, source_name, size, name):
        source = self.get_source(source_name)
        meta = ThumbnailMeta.objects.create(source=source, size=size, name=name)
        return ImageMeta(source_name, meta.name, meta.size)

    def delete_thumbnail(self, source_name, size):
        ThumbnailMeta.objects.filter(source__name=source_name, size=size).delete()

    def flush_thumbnails(self, source_name):
        ThumbnailMeta.objects.filter(source__name=source_name).delete()


class RedisBackend(BaseBackend):

    def __init__(self):
        host = conf.METADATA.get('host', 'localhost')
        port = conf.METADATA.get('port', 6379)
        password = conf.METADATA.get('password', None)
        db = conf.METADATA.get('db', 0)
        prefix = conf.METADATA.get('PREFIX', 'djthumbs')
        self.prefix = prefix + ":"
        if not StrictRedis:
            msg = "Could not import Redis. Please install 'redis' extra."
            raise ImportError(msg)
        self.redis = StrictRedis(host=host, port=port, password=password, db=db)

    def get_source_key(self, name):
        return "%ssources:%s" % (self.prefix, name)

    def get_thumbnail_key(self, name):
        return "%sthumbnails:%s" % (self.prefix, name)

    def add_source(self, name):
        self.redis.hset(self.get_source_key(name), name, name)
        return name

    def get_source(self, name):
        return compat.as_text(self.redis.hget(self.get_source_key(name), name))

    def delete_source(self, name):
        return self.redis.hdel(self.get_source_key(name), name)

    def get_thumbnails(self, name):
        metas = self.redis.hgetall(self.get_thumbnail_key(name))
        return [ImageMeta(name, thumbnail_name, size) for size, thumbnail_name in metas.items()]

    def get_thumbnail(self, source_name, size):
        name = self.redis.hget(self.get_thumbnail_key(source_name), size)
        if name:
            return ImageMeta(source_name, name, size)
        return None

    def add_thumbnail(self, source_name, size, name):
        self.redis.hset(self.get_thumbnail_key(source_name), size, name)
        return ImageMeta(source_name, name, size)

    def delete_thumbnail(self, source_name, size):
        self.redis.hdel(self.get_thumbnail_key(source_name), size)

    def flush_thumbnails(self, source_name):
        self.redis.delete(self.get_thumbnail_key(source_name))



from django.test import TestCase
from thumbnails import conf


class SettingsTest(TestCase):

    def test_load_settings(self):
        self.assertNotEqual(conf.METADATA, None)
        self.assertNotEqual(conf.STORAGE, None)
        self.assertNotEqual(conf.THUMBNAILS, {})
        self.assertNotEqual(conf.SIZES, {})



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Source'
        db.create_table(u'thumbnails_source', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(unique=True, max_length=255)),
        ))
        db.send_create_signal(u'thumbnails', ['Source'])

        # Adding model 'ThumbnailMeta'
        db.create_table(u'thumbnails_thumbnailmeta', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('source', self.gf('django.db.models.fields.related.ForeignKey')(related_name='thumbnails', to=orm['thumbnails.Source'])),
            ('size', self.gf('django.db.models.fields.CharField')(max_length=64)),
            ('name', self.gf('django.db.models.fields.CharField')(unique=True, max_length=255)),
        ))
        db.send_create_signal(u'thumbnails', ['ThumbnailMeta'])

        # Adding unique constraint on 'ThumbnailMeta', fields ['source', 'size']
        db.create_unique(u'thumbnails_thumbnailmeta', ['source_id', 'size'])


    def backwards(self, orm):
        # Removing unique constraint on 'ThumbnailMeta', fields ['source', 'size']
        db.delete_unique(u'thumbnails_thumbnailmeta', ['source_id', 'size'])

        # Deleting model 'Source'
        db.delete_table(u'thumbnails_source')

        # Deleting model 'ThumbnailMeta'
        db.delete_table(u'thumbnails_thumbnailmeta')


    models = {
        u'thumbnails.source': {
            'Meta': {'object_name': 'Source'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255'})
        },
        u'thumbnails.thumbnailmeta': {
            'Meta': {'unique_together': "(('source', 'size'),)", 'object_name': 'ThumbnailMeta'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255'}),
            'size': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            'source': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'thumbnails'", 'to': u"orm['thumbnails.Source']"})
        }
    }

    complete_apps = ['thumbnails']


from django.db.models.fields.files import ImageFieldFile

from . import conf, images
from .backends.storage import get_backend
from .images import Thumbnail, FallbackImage
from .metadata import get_path


class SourceImage(ImageFieldFile):

    def __init__(self, name):
        self.name = name


class ThumbnailedImageFile(ImageFieldFile):

    def __init__(self, instance, field, name, **kwargs):
        super(ThumbnailedImageFile, self).__init__(instance, field, name, **kwargs)
        self.metadata_backend = field.metadata_backend
        self.thumbnails = ThumbnailManager(
            metadata_backend=self.metadata_backend,
            storage=self.storage,
            source_image=self
        )

    def save(self, name, content, save=True):
        thumbnail = super(ThumbnailedImageFile, self).save(name, content, save)
        self.metadata_backend.add_source(self.name)
        return thumbnail

    def delete(self, with_thumbnails=True, save=True):
        if not with_thumbnails:
            super().delete(save=save)
            return

        self.thumbnails.delete_all()
        super().delete(save=save)


class ThumbnailManager(object):
    """A class that manages creation and retrieval of thumbnails."""

    def __init__(self, metadata_backend, storage, source_image):
        self.metadata_backend = metadata_backend
        self.storage = storage
        self.source_image = source_image
        self._thumbnails = None

    def __getattr__(self, name):
        if name in conf.SIZES.keys():
            if not self.source_image:
                fallback_image_url = conf.SIZES[name].get('FALLBACK_IMAGE_URL')
                if fallback_image_url:
                    return FallbackImage(fallback_image_url)
                else:
                    return Thumbnail(metadata=None, storage=self.storage)
            return self.get(name)
        else:
            raise AttributeError("'%s' has no attribute '%s'" % (self, name))

    def _refresh_cache(self):
        """Populate self._thumbnails."""
        self._thumbnails = {}
        metadatas = self.metadata_backend.get_thumbnails(self.source_image.name)
        for metadata in metadatas:
            self._thumbnails[metadata.size] = Thumbnail(metadata=metadata, storage=self.storage)

    def all(self):
        """
        Return all thumbnails in a dict format.
        """
        if self._thumbnails is not None:
            return self._thumbnails
        self._refresh_cache()
        return self._thumbnails

    def get(self, size, create=True):
        """
        Returns a Thumbnail instance.
        First check whether thumbnail is already cached. If it doesn't:
        1. Try to fetch the thumbnail
        2. Create thumbnail if it's not present
        3. Cache the thumbnail for future use
        """
        if self._thumbnails is None:
            self._refresh_cache()

        thumbnail = self._thumbnails.get(size)

        if thumbnail is None:
            thumbnail = images.get(self.source_image.name, size,
                                   self.metadata_backend, self.storage)

            if thumbnail is None:
                thumbnail = self.create(size)

            self._thumbnails[size] = thumbnail

        return thumbnail

    def create(self, size):
        """
        Creates and return a thumbnail of a given size.
        """
        thumbnail = images.create(self.source_image.name, size,
                                  self.metadata_backend, self.storage)
        return thumbnail

    def delete(self, size):
        """
        Deletes a thumbnail of a given size
        """
        images.delete(self.source_image.name, size,
                      self.metadata_backend, self.storage)

        if self._thumbnails is not None and self._thumbnails.get(size):
            del(self._thumbnails[size])

    def delete_all(self):
        self._thumbnails = None

        for size, thumbnail in self.all().items():
            self.storage.delete(thumbnail.name)

        self.metadata_backend.flush_thumbnails(self.source_image.name)


def exists(source_name, size=None):
    path = get_path(source_name, size)
    if path is not None:
        return get_backend().exists(path)
    else:
        return False


def delete(source_name, size=None):
    path = get_path(source_name, size)
    return get_backend().delete(path)



import os

from django.utils.encoding import force_str

from . import conf
from . import backends
from . import post_processors
from . import processors


class Thumbnail(object):
    """
    An object that contains relevant information about a thumbnailed image.
    """

    def __init__(self, metadata, storage):
        self.metadata = metadata
        self.storage = storage
        self.name = getattr(metadata, 'name', None)

    def __str__(self):
        return force_str(self.name or '')

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self.name or "None")

    def __eq__(self, other):
        try:
            return self.__dict__ == other.__dict__
        except AttributeError:
            return False

    def __bool__(self):
        return bool(self.name)

    def check_metadata(self):
        if self.metadata is None:
            raise ValueError('Thumbnail has no source file')

    @property
    def size(self):
        self.check_metadata()
        return self.metadata.size

    @property
    def url(self):
        self.check_metadata()
        return self.storage.url(self.name)


class FallbackImage(object):

    def __init__(self, image_url):
        self.image_url = image_url

    @property
    def url(self):
        return self.image_url


def get_thumbnail_name(source_name, size):
    name, extension = os.path.splitext(source_name)
    if conf.SIZES[size].get("FORMAT"):
        extension = ".{}".format(conf.SIZES[size]['FORMAT'])
    filename = "%s_%s%s" % (name, size, extension)
    return os.path.join(conf.BASE_DIR, filename)


def create(source_name, size, metadata_backend=None, storage_backend=None):
    """
    Creates a thumbnail file and its relevant metadata. Returns a
    Thumbnail instance.
    """
    if storage_backend is None:
        storage_backend = backends.storage.get_backend()
    if metadata_backend is None:
        metadata_backend = backends.metadata.get_backend()

    thumbnail_file = processors.process(storage_backend.open(source_name), size)
    thumbnail_file = post_processors.process(thumbnail_file, size)

    return save(source_name, size, metadata_backend, storage_backend, thumbnail_file)


def save(source_name, size, metadata_backend, storage_backend, image_file):
    name = get_thumbnail_name(source_name, size)
    name = storage_backend.save(name, image_file)

    metadata = metadata_backend.add_thumbnail(source_name, size, name)
    return Thumbnail(metadata=metadata, storage=storage_backend)


def get(source_name, size, metadata_backend=None, storage_backend=None):
    """
    Returns a Thumbnail instance, or None if thumbnail does not yet exist.
    """
    if storage_backend is None:
        storage_backend = backends.storage.get_backend()
    if metadata_backend is None:
        metadata_backend = backends.metadata.get_backend()

    metadata = metadata_backend.get_thumbnail(source_name, size)
    if metadata is None:
        return None
    else:
        return Thumbnail(metadata=metadata, storage=storage_backend)


def delete(source_name, size, metadata_backend=None, storage_backend=None):
    """
    Deletes a thumbnail file and its relevant metadata.
    """
    if storage_backend is None:
        storage_backend = backends.storage.get_backend()
    if metadata_backend is None:
        metadata_backend = backends.metadata.get_backend()

    thumbnail = get(source_name, size, metadata_backend, storage_backend)
    if thumbnail:
        storage_backend.delete(thumbnail.name)

    metadata_backend.delete_thumbnail(source_name, size)



# -*- coding: utf-8 -*-


DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
    },
}


INSTALLED_APPS = (
    'thumbnails',
    'thumbnails.tests',
)

SECRET_KEY = 'a'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

THUMBNAILS = {
    'METADATA': {
        'PREFIX': 'djthumbs-test',
        'BACKEND': 'thumbnails.backends.metadata.DatabaseBackend',
    },
    'STORAGE': {
        'BACKEND': 'thumbnails.tests.storage.TemporaryStorage'
    },
    'BASE_DIR': 'thumbs',
    'SIZES': {
        'small': {
            'PROCESSORS': [
                {'PATH': 'thumbnails.processors.resize', 'width': 10, 'height': 10}
            ]
        },
        'default': {
            'FALLBACK_IMAGE_URL': 'thumbnails/tests/tests.png',
            'PROCESSORS': [
                {'PATH': 'thumbnails.processors.resize', 'width': 20, 'height': 20},
                {'PATH': 'thumbnails.processors.flip', 'direction': 'horizontal'}
            ],
            'POST_PROCESSORS': [
                {'PATH': 'thumbnails.post_processors.optimize', 'png_command': 'optipng %(filename)s'},
            ]
        },
        'large': {
            'PROCESSORS': [
                {'PATH': 'thumbnails.processors.resize', 'width': 80, 'height': 80},
                {'PATH': 'thumbnails.processors.rotate', 'degrees': 45},
                {'PATH': 'thumbnails.processors.crop', 'width': 80, 'height': 80, 'center': ('50%,50%')}
            ]
        },
        'source': {
            'PROCESSORS': [
                {'PATH': 'thumbnails.processors.resize', 'width': 90, 'height': 90}
            ]
        },
        'source_with_format': {
            'PROCESSORS': [
                {'PATH': 'thumbnails.processors.resize', 'width': 90, 'height': 90}
            ],
            'FORMAT': 'webp',
        }
    }
}

MIDDLEWARE_CLASSES = ()



from django.db import models


class Source(models.Model):
    name = models.CharField(unique=True, max_length=255)

    class Meta:
        app_label = 'thumbnails'


class ThumbnailMeta(models.Model):
    source = models.ForeignKey(Source, related_name='thumbnails',
                               on_delete=models.CASCADE)
    size = models.CharField(max_length=64)
    name = models.CharField(unique=True, max_length=255)

    class Meta:
        unique_together = ('source', 'size')
        app_label = 'thumbnails'



from django.test import TestCase
from django.core.files import File
from django.core.management import call_command

from .models import TestModel
from thumbnails.backends import metadata


class CommandTest(TestCase):

    def test_delete_thumbnails(self):
        metadata_backend = metadata.get_backend()
        get_thumbnail = metadata_backend.get_thumbnail
        instance = TestModel.objects.create()

        with open('thumbnails/tests/tests.png', 'rb') as image_file:
            instance.avatar = File(image_file)
            instance.save()

        with open('thumbnails/tests/tests.png', 'rb') as image_file:
            instance.profile_picture = File(image_file)
            instance.save()

        instance = TestModel.objects.get(id=instance.id)
        instance.avatar.thumbnails.create(size='small')
        instance.avatar.thumbnails.create(size='large')
        instance.profile_picture.thumbnails.create(size='small')
        instance.profile_picture.thumbnails.create(size='large')

        self.assertTrue(get_thumbnail(instance.avatar.name, 'small'))
        self.assertTrue(get_thumbnail(instance.avatar.name, 'large'))
        self.assertTrue(get_thumbnail(instance.profile_picture.name, 'small'))
        self.assertTrue(get_thumbnail(instance.profile_picture.name, 'large'))

        call_command('delete_thumbnails', path_to_model='tests.TestModel',
                     field_name='avatar', size='small')

        # Ensure only the relevant thumbnails are deleted
        self.assertFalse(get_thumbnail(instance.avatar.name, 'small'))
        self.assertTrue(get_thumbnail(instance.avatar.name, 'large'))
        self.assertTrue(get_thumbnail(instance.profile_picture.name, 'small'))
        self.assertTrue(get_thumbnail(instance.profile_picture.name, 'large'))



import os

from da_vinci import images
from django.test import TestCase
from django.core.files import File
from PIL import Image

from thumbnails.processors import add_watermark

from .storage import TemporaryStorage


class ProcessorsTest(TestCase):

    def test_add_watermark(self):
        image_path = 'thumbnails/tests/tests.png'
        watermark_path = os.path.join(TemporaryStorage().temporary_location, "test_watermark.png") 
        original_image_path = os.path.join(TemporaryStorage().temporary_location, "test_bg.png") 

        # different size
        original_image_file = images.from_file(image_path)
        original_image = original_image_file.get_pil_image()
        converted_image = original_image.resize((300, 300))
        converted_image.save(original_image_path)
        original_image_file.set_pil_image(converted_image)

        watermark_image = Image.open(image_path)
        converted_watermark = watermark_image.resize((200, 200))
        converted_watermark.save(watermark_path)

        kwargs = {
            "watermark_path": watermark_path
        }
        self.assertRaises(ValueError, add_watermark, original_image_file, **kwargs)

        converted_image = original_image.resize((200, 200))
        converted_image.save(original_image_path)
        original_image_file.set_pil_image(converted_image)

        # not rgba
        converted_watermark = converted_watermark.convert("RGB")
        converted_watermark.save(watermark_path)

        self.assertRaises(ValueError, add_watermark, original_image_file, **kwargs)

        converted_watermark = converted_watermark.convert("RGBA")
        converted_watermark.save(watermark_path)

        # no error raised
        add_watermark(original_image_file, **kwargs)



from copy import deepcopy

try:
    import importlib
except ImportError:
    from django.utils import importlib


def import_attribute(name):
    """
    Return an attribute from a dotted path name (e.g. "path.to.func").
    Copied from nvie's rq https://github.com/nvie/rq/blob/master/rq/utils.py
    """
    if hasattr(name, '__call__'):
        return name
    module_name, attribute = name.rsplit('.', 1)
    module = importlib.import_module(module_name)
    return getattr(module, attribute)


def parse_processors(processor_definition):
    """
    Returns a dictionary that contains the imported processors and
    kwargs. For example, passing in:

    processors = [
        {'processor': 'thumbnails.processors.resize', 'width': 10, 'height': 10},
        {'processor': 'thumbnails.processors.crop', 'width': 10, 'height': 10},
    ]

    Would return:

    [
        {'processor': resize_function, kwargs: {'width': 10, 'height': 10}}
        {'processor': crop_function, kwargs: {'width': 10, 'height': 10}}
    ]
    """
    parsed_processors = []
    for processor in processor_definition:
        processor_function = import_attribute(processor['PATH'])
        kwargs = deepcopy(processor)
        kwargs.pop('PATH')
        parsed_processors.append({
            'processor': processor_function,
            'kwargs': kwargs
        })

    return parsed_processors



import os

from django.core.files import File
from django.template import Context, Template
from django.test import TestCase

from thumbnails import conf
from thumbnails.backends.metadata import RedisBackend
from thumbnails.fields import fetch_thumbnails
from thumbnails.files import Thumbnail, FallbackImage

from .models import TestModel, TestPregeneratedSizesModel
from .storage import TemporaryStorage2, TemporaryStorage


class ImageFieldTest(TestCase):

    def setUp(self):
        self.instance = TestModel.objects.create()
        # Resize Without Image Convert
        with open('thumbnails/tests/tests.png', 'rb') as image_file:
            self.instance.avatar = File(image_file)
            self.instance.save()
        self.avatar_folder = \
            os.path.join(self.instance.avatar.storage.temporary_location, conf.BASE_DIR, 'avatars')
        self.basename = os.path.basename(self.instance.avatar.path)
        self.filename, self.ext = os.path.splitext(self.basename)

        # Resize with Image Convert
        with open('thumbnails/tests/tests.png', 'rb') as image_file:
            self.instance.card_identity_picture = File(image_file)
            self.instance.save()

        self.identity_card_folder = \
            os.path.join(self.instance.card_identity_picture.storage.temporary_location, conf.BASE_DIR, 'identity_card')
        self.identity_picture_basename = os.path.basename(self.instance.card_identity_picture.path)
        self.identity_filename, self.identity_ext = os.path.splitext(self.identity_picture_basename)

    def tearDown(self):
        self.instance.avatar.storage.delete_temporary_storage()
        self.instance.card_identity_picture.storage.delete_temporary_storage()

        super(ImageFieldTest, self).tearDown()

    def test_image_field(self):

        # 1. Test for thumbnail creation
        self.assertFalse(os.path.isfile(os.path.join(self.avatar_folder, self.filename + '_small' + self.ext)))
        thumb = self.instance.avatar.thumbnails.create(size='small')
        self.assertTrue(os.path.isfile(os.path.join(self.avatar_folder, self.filename + '_small' + self.ext)))

        # Make sure the returned thumbnail is of thumbnail class, not metadata
        self.assertTrue(isinstance(thumb, Thumbnail))
        # 2. Test for getting thumbnail
        self.assertEqual(thumb, self.instance.avatar.thumbnails.get(size='small'))

        # 3. Test for thumbnail deletion
        self.assertTrue(os.path.isfile(os.path.join(self.avatar_folder, self.filename + '_small' + self.ext)))
        self.instance.avatar.thumbnails.delete(size='small')
        self.assertFalse(os.path.isfile(os.path.join(self.avatar_folder, self.filename + '_small' + self.ext)))

        # Test convert png image to webp image, ImageField with resize and pregenerated sizes
        self.assertEqual(self.identity_ext, '.webp')

        self.assertIsInstance(thumb.storage, TemporaryStorage)

        # After convert to webp, make sure resize can be running as normal
        # 1. Test for thumbnail creation
        self.assertFalse(os.path.isfile(os.path.join(self.identity_card_folder, self.identity_filename + '_small' + self.identity_ext)))
        thumb = self.instance.card_identity_picture.thumbnails.create(size='small')
        self.assertTrue(os.path.isfile(os.path.join(self.identity_card_folder, self.identity_filename + '_small' + self.identity_ext)))

        # 2. Test for getting thumbnail
        self.assertEqual(thumb, self.instance.card_identity_picture.thumbnails.get(size='small'))

        # 3. Test for thumbnail deletion
        self.assertTrue(os.path.isfile(os.path.join(self.identity_card_folder, self.identity_filename + '_small' + self.identity_ext)))
        self.instance.card_identity_picture.thumbnails.delete(size='small')
        self.assertFalse(os.path.isfile(os.path.join(self.identity_card_folder, self.identity_filename + '_small' + self.identity_ext)))

        self.assertIsInstance(thumb.storage, TemporaryStorage2)

    def test_thumbnail_field(self):
        # Make sure ThumbnailManager return the correct thumbnail
        self.assertTrue(self.instance.avatar.thumbnails.small, Thumbnail)

        self.assertEqual(os.path.basename(self.instance.avatar.thumbnails.small.name),
                         self.filename + '_small' + self.ext)
        self.assertTrue(self.instance.avatar.thumbnails.default, Thumbnail)
        self.assertEqual(os.path.basename(self.instance.avatar.thumbnails.default.name),
                         self.filename + '_default' + self.ext)
        self.assertTrue(self.instance.avatar.thumbnails.large, Thumbnail)
        self.assertEqual(os.path.basename(self.instance.avatar.thumbnails.large.name),
                         self.filename + '_large' + self.ext)

        # Test for name clashing with another file with the same name
        self.instance.avatar.thumbnails.delete('large')
        open(os.path.join(self.avatar_folder, 'tests_large.png'), 'w').close()
        self.instance.avatar.thumbnails.create('large')

        # Due to uuid4 for file name, this should not clash
        self.assertNotEqual(os.path.basename(self.instance.avatar.thumbnails.large.name),
                            'tests_large.png')

        # Create Thumbnail with source_with_format, formated image with 'webp'
        self.assertEqual(os.path.basename(self.instance.avatar.thumbnails.source_with_format.name),
                         self.filename + '_source_with_format' + '.webp')

        # Make sure ThumbnailManager return the correct thumbnail with Image Convert (webp)
        self.assertTrue(self.instance.card_identity_picture.thumbnails.small, Thumbnail)
        self.assertEqual(os.path.basename(self.instance.card_identity_picture.thumbnails.small.name),
                         self.identity_filename + '_small' + self.identity_ext)
        self.assertTrue(self.instance.card_identity_picture.thumbnails.default, Thumbnail)
        self.assertEqual(os.path.basename(self.instance.card_identity_picture.thumbnails.default.name),
                         self.identity_filename + '_default' + self.identity_ext)
        self.assertTrue(self.instance.card_identity_picture.thumbnails.large, Thumbnail)
        self.assertEqual(os.path.basename(self.instance.card_identity_picture.thumbnails.large.name),
                         self.identity_filename + '_large' + self.identity_ext)

        # Test for name clashing with another file with the same name
        self.instance.card_identity_picture.thumbnails.delete('large')
        open(os.path.join(self.identity_card_folder, 'tests_large.webp'), 'w').close()
        self.instance.card_identity_picture.thumbnails.create('large')

        # Due to uuid4 for file name, this should not clash
        self.assertNotEqual(os.path.basename(self.instance.card_identity_picture.thumbnails.large.name),
                            'tests_large.webp')

    def test_thumbnails_cache(self):

        # No thumbnails should be cached
        self.assertEqual(self.instance.avatar.thumbnails._thumbnails, None)

        # Thumbnail with size `default` created and populated to the cache
        self.assertEqual(self.instance.avatar.thumbnails.default, self.instance.avatar.thumbnails.all().get('default'))
        self.assertEqual(len(self.instance.avatar.thumbnails._thumbnails), 1)

        # Creating thumbnail should populate the cache correctly
        large_thumb = self.instance.avatar.thumbnails.large
        self.assertEqual(large_thumb, self.instance.avatar.thumbnails.all().get('large'))
        self.assertEqual(len(self.instance.avatar.thumbnails._thumbnails), 2)

        # Creating thumbnail should populate the cache correctly (webp)
        source_with_format_thumb = self.instance.avatar.thumbnails.source_with_format
        self.assertEqual(source_with_format_thumb, self.instance.avatar.thumbnails.all().get('source_with_format'))
        # Making sure the file format correctly
        self.assertEqual(os.path.basename(self.instance.avatar.thumbnails.source_with_format.name),
                         self.filename + '_source_with_format' + '.webp')
        self.assertEqual(len(self.instance.avatar.thumbnails._thumbnails), 3)

        # Should also work on deletion
        self.instance.avatar.thumbnails.delete('large')
        self.assertEqual(self.instance.avatar.thumbnails.all().get('large'), None)
        self.assertEqual(len(self.instance.avatar.thumbnails._thumbnails), 2)

        # Should also work on deletion (webp)
        self.instance.avatar.thumbnails.delete('source_with_format')
        self.assertEqual(self.instance.avatar.thumbnails.all().get('source_with_format'), None)
        self.assertEqual(len(self.instance.avatar.thumbnails._thumbnails), 1)

        # Once cached, it should not hit backend on other call.
        with self.assertNumQueries(0):
            self.instance.avatar.thumbnails.default
            self.instance.avatar.thumbnails.all()['default']
            self.instance.avatar.thumbnails.get('default')
            self.instance.avatar.thumbnails.get('default')

        # With Image resize & convert
        # No thumbnails should be cached
        self.assertEqual(self.instance.card_identity_picture.thumbnails._thumbnails, None)

        # Thumbnail with size `default` created and populated to the cache
        self.assertEqual(self.instance.card_identity_picture.thumbnails.default, self.instance.card_identity_picture.thumbnails.all().get('default'))
        self.assertEqual(len(self.instance.card_identity_picture.thumbnails._thumbnails), 1)

        # Creating thumbnail should populate the cache correctly
        large_thumb = self.instance.card_identity_picture.thumbnails.large
        self.assertEqual(large_thumb, self.instance.card_identity_picture.thumbnails.all().get('large'))
        self.assertEqual(len(self.instance.card_identity_picture.thumbnails._thumbnails), 2)

        # Should also work on deletion
        self.instance.card_identity_picture.thumbnails.delete('large')
        self.assertEqual(self.instance.card_identity_picture.thumbnails.all().get('large'), None)
        self.assertEqual(len(self.instance.card_identity_picture.thumbnails._thumbnails), 1)

        # Once cached, it should not hit backend on other call.
        with self.assertNumQueries(0):
            self.instance.card_identity_picture.thumbnails.default
            self.instance.card_identity_picture.thumbnails.all()['default']
            self.instance.card_identity_picture.thumbnails.get('default')
            self.instance.card_identity_picture.thumbnails.get('default')

    def test_django_template(self):
        template = Template("Test render {{ image.thumbnails.large.url }} ")
        context = Context({"image": self.instance.avatar})
        result = template.render(context)
        expected_path = os.path.join('thumbs', 'avatars', self.filename + "_large" + self.ext)
        self.assertIn(expected_path, result)

    def test_thumbnails_without_source(self):
        template = Template("Test render {{ image.thumbnails.large.url }} ")

        test_model = TestModel.objects.create()
        test_model.avatar = None
        test_model.save()
        context = Context({"image": test_model.avatar})
        self.assertRaises(ValueError, template.render, context)

        test_model = TestModel.objects.create()
        test_model.avatar = ''
        test_model.save()
        context = Context({"image": test_model.avatar})
        self.assertRaises(ValueError, template.render, context)

        # If source_image is None, ValueError should be raised when calling functions
        # AttributeError must be raised if ThumbnailManager is called with non existent size
        # This applies when no default picture is defined
        self.assertFalse(test_model.avatar)
        self.assertRaises(ValueError, getattr, test_model.avatar.thumbnails.large, 'url')
        self.assertRaises(AttributeError, getattr, test_model.avatar.thumbnails, 'lrge')

    def test_fallback_image(self):

        # We have defined FALLBACK_IMAGE_URL for size ``default``
        self.assertFalse(self.instance.profile_picture)
        self.assertTrue(isinstance(self.instance.profile_picture.thumbnails.default, FallbackImage))

        # No errors should be raised when url attribute
        self.instance.profile_picture.thumbnails.default.url

        # No FALLBACK_IMAGE_URL defined for size ``large``
        self.assertFalse(self.instance.profile_picture)
        self.assertTrue(isinstance(self.instance.profile_picture.thumbnails.large, Thumbnail))

        # Error should be raised when calling url attribute
        self.assertRaises(ValueError, getattr, self.instance.profile_picture.thumbnails.large, 'url')
        self.assertRaises(ValueError, getattr, self.instance.profile_picture.thumbnails.large, 'size')

    def test_resize_source_to_none(self):
        # Make sure that fields without resize_source_to does not raise error on save
        with open('thumbnails/tests/tests.png', 'rb') as image_file:
            self.instance.profile_picture = File(image_file)
            self.instance.save()

    def test_fetch_non_redis(self):
        test_objc = TestModel.objects.create()
        with open('thumbnails/tests/tests.png', 'rb') as image_file:
            test_objc.avatar = File(image_file)
            test_objc.save()

        # create all thumbnails
        objects = TestModel.objects.all()
        images = []
        for obj in objects:
            for size in conf.SIZES:
                obj.avatar.thumbnails.get(size)
            images.append(obj.avatar)

        # reset _thumbnails
        for image in images:
            image.thumbnails._thumbnails = {}

        # default backend(thumbnails.backends.metadata.DatabaseBackend) is not supported
        self.assertRaises(NotImplementedError, fetch_thumbnails, images)

    def test_fetch_redis_backend(self):
        TestModel.objects.all().delete()

        for i in range(1, 10):
            test_objc = TestModel.objects.create()

            with open('thumbnails/tests/tests.png', 'rb') as image_file:
                test_objc.avatar = File(image_file)
                test_objc.save()

        images = []
        objects = TestModel.objects.all()
        # test pre-created thumbnails
        for obj in objects:
            obj.avatar.thumbnails.metadata_backend = RedisBackend()
            images.append(obj.avatar)

        # make sure cache is still empty, so it can be generated freshly
        fetch_thumbnails(images)
        for image in images:
            self.assertEqual(image.thumbnails._thumbnails, {})

        # create all thumbnails
        for image in images:
            for size in conf.SIZES:
                image.thumbnails.get(size)

        # reset _thumbnails
        for image in images:
            image.thumbnails._thumbnails = {}

        fetch_thumbnails(images)
        for image in images:
            thumbnails = image.thumbnails
            sizes = [size for size in thumbnails._thumbnails.keys()]
            for size in sizes:
                # Make sure all thumbnail sizes have the right value
                self.assertEqual(thumbnails._thumbnails[size].name,
                                 thumbnails.get(size).name)
            self.assertEqual(set(sizes), set(conf.SIZES))

        # Fetch for format image run correctly
        fetch_thumbnails(images, ['source_with_format'])
        for image in images:
            thumbnails = image.thumbnails
            sizes = [size for size in thumbnails._thumbnails.keys()]
            self.assertEqual(thumbnails._thumbnails['source_with_format'].name,
                             thumbnails.get('source_with_format').name)
            self.assertEqual(sizes, ['source_with_format'])
            self.assertEqual(os.path.splitext(thumbnails._thumbnails['source_with_format'].name)[1],
                             '.webp')

    def test_populate_redis_backend_with_size(self):
        TestModel.objects.all().delete()

        for i in range(1, 10):
            test_objc = TestModel.objects.create()

            with open('thumbnails/tests/tests.png', 'rb') as image_file:
                test_objc.avatar = File(image_file)
                test_objc.save()

        # create all thumbnails
        objects = TestModel.objects.all()
        images = []
        for obj in objects:
            obj.avatar.thumbnails.metadata_backend = RedisBackend()
            for size in conf.SIZES:
                obj.avatar.thumbnails.get(size)
            images.append(obj.avatar)

        # reset _thumbnails
        for image in images:
            image.thumbnails._thumbnails = {}

        # Adding format image run correctly
        fetch_thumbnails(images, ['small', 'large', 'source_with_format'])
        for image in images:
            thumbnails = image.thumbnails
            sizes = [size for size in thumbnails._thumbnails.keys()]
            for size in sizes:
                # Make sure all thumbnail sizes have the right value
                self.assertEqual(thumbnails._thumbnails[size].name,
                                 thumbnails.get(size).name)
            self.assertEqual(set(sizes), set(['small', 'large', 'source_with_format']))


class PregeneratedFilesTest(TestCase):
    def setUp(self):
        self.instance = TestPregeneratedSizesModel.objects.create()

    def tearDown(self):
        if self.instance.logo:
            self.instance.logo.storage.delete_temporary_storage()
        if self.instance.photo:
            self.instance.photo.storage.delete_temporary_storage()

        super(PregeneratedFilesTest, self).tearDown()

    def test_image_field(self):
        # Test for thumbanils with pregenerated sizes
        with open('thumbnails/tests/tests.png', 'rb') as image_file:
            self.instance.logo = File(image_file)
            self.instance.save()

        logo_folder = \
            os.path.join(self.instance.logo.storage.temporary_location, conf.BASE_DIR, 'logos')
        logo_basename = os.path.basename(self.instance.logo.path)
        logo_filename, logo_ext = os.path.splitext(logo_basename)

        pregenerated_files = os.listdir(logo_folder)
        self.assertEqual(len(pregenerated_files), 2)
        for size in ['_large', '_small']:
            self.assertIn(logo_filename + size + logo_ext, pregenerated_files)

        # Test for thumbnails with resource to and pregenerated files
        with open('thumbnails/tests/tests.png', 'rb') as image_file:
            self.instance.photo = File(image_file)
            self.instance.save()

        photo_folder = \
            os.path.join(self.instance.photo.storage.temporary_location, conf.BASE_DIR, 'photos')
        photo_basename = os.path.basename(self.instance.photo.path)
        photo_filename, photo_ext = os.path.splitext(photo_basename)

        pregenerated_files = os.listdir(photo_folder)
        self.assertEqual(len(pregenerated_files), 3)
        for size in ['_large', '_default', '_source_with_format']:
            self.assertIn(photo_filename + size + photo_ext, pregenerated_files)



import os

from django.core.files import File
from django.test import TestCase

from unittest.mock import patch
from thumbnails import images, conf

from .models import TestModel


class ImageTest(TestCase):

    def setUp(self):
        self.instance = TestModel.objects.create()
        with open('thumbnails/tests/tests.png', 'rb') as image_file:
            self.instance.avatar = File(image_file)
            self.instance.save()

        self.basename = os.path.basename(self.instance.avatar.path)
        self.filename, self.ext = os.path.splitext(self.basename)
        self.source_name = os.path.join('avatars', self.filename + self.ext)

        self.storage_backend = self.instance.avatar.storage
        self.metadata_backend = self.instance.avatar.metadata_backend

    def tearDown(self):
        self.instance.avatar.storage.delete_temporary_storage()
        super(ImageTest, self).tearDown()

    def test_create(self):
        """
        Ensure that ``create`` properly generates the thumbnail
        and its relevant metadata.
        """
        thumbnail_name = images.get_thumbnail_name(self.source_name, 'small')

        self.assertEqual(
            self.metadata_backend.get_thumbnail(self.source_name, 'small'),
            None
        )
        self.assertFalse(self.storage_backend.exists(thumbnail_name))

        thumbnail = images.create(self.source_name, 'small',
                                  self.metadata_backend, self.storage_backend)
        self.assertTrue(self.storage_backend.exists(thumbnail.name))
        self.assertNotEqual(
            self.metadata_backend.get_thumbnail(self.source_name, 'small'),
            None
        )

    def test_get(self):
        """
        Ensure that ``get`` works properly.
        """
        self.assertEqual(
            images.get(self.source_name, 'default',
                       self.metadata_backend, self.storage_backend),
            None
        )
        thumbnail = images.create(self.source_name, 'default',
                                  self.metadata_backend, self.storage_backend)

        self.assertEqual(
            images.get(self.source_name, 'default',
                       self.metadata_backend, self.storage_backend),
            thumbnail
        )

    def test_get_thumbnail_name(self):
        # test with size in settings
        expected_basename = "%s_%s%s" % (self.filename, "small", self.ext)
        basename = os.path.basename(images.get_thumbnail_name(self.instance.avatar.name, "small"))
        self.assertEqual(basename, expected_basename)

        # test with format
        expected_basename = "%s_%s%s" % (self.filename, "source_with_format", ".webp")
        basename = os.path.basename(images.get_thumbnail_name(self.instance.avatar.name, "source_with_format"))
        self.assertEqual(basename, expected_basename)

    def test_delete(self):
        """
        Ensure that ``delete`` works properly.
        """
        thumbnail = images.create(self.source_name, 'small',
                                  self.metadata_backend, self.storage_backend)

        images.delete(self.source_name, 'small',
                      self.metadata_backend, self.storage_backend)
        self.assertFalse(self.storage_backend.exists(thumbnail.name))
        self.assertEqual(
            self.metadata_backend.get_thumbnail(self.source_name, 'small'),
            None
        )

        # key with format is not defined in settings anymore
        thumbnail = images.create(self.source_name, 'source_with_format',
                                  self.metadata_backend, self.storage_backend)

        thumbnail_size_data = {
            'small': {
                'PROCESSORS': [
                    {'PATH': 'thumbnails.processors.resize', 'width': 10, 'height': 10}
                ]
            },
        }
        with patch('thumbnails.images.conf.SIZES', thumbnail_size_data):
            # ensure error not happening, and its files and metadata are still removed
            images.delete(self.source_name, 'source_with_format',
                          self.metadata_backend, self.storage_backend)
            self.assertFalse(self.storage_backend.exists(thumbnail.name))
            self.assertIsNone(self.metadata_backend.get_thumbnail(self.source_name, 'source_with_format'))

    def test_delete_with_no_thumbnails_file(self):
        avatar_path = self.instance.avatar.path
        avatar_folder = os.path.join(self.instance.avatar.storage.temporary_location, conf.BASE_DIR, 'avatars')
        self.assertTrue(os.path.exists(avatar_path))

        self.instance.avatar.thumbnails.small

        size = 'small'
        thumbnail_name = os.path.basename(images.get_thumbnail_name(self.instance.avatar.name, size))
        thumbnail_path = os.path.join(avatar_folder, thumbnail_name)
        self.assertTrue(os.path.exists(thumbnail_path))

        # test delete non existence thumbnail file, should not raise error
        os.remove(thumbnail_path)
        self.assertFalse(os.path.exists(thumbnail_path))

        images.delete(self.source_name, size,
                      self.metadata_backend, self.storage_backend)

        # thumbnails and their metadata are also deleted
        self.assertEqual(len(os.listdir(avatar_folder)), 0)
        self.assertIsNone(self.metadata_backend.get_thumbnail(self.source_name, size))



from redis import StrictRedis
from django.test import TestCase

from thumbnails.backends.metadata import ImageMeta
from thumbnails.models import Source, ThumbnailMeta
from thumbnails.backends.metadata import DatabaseBackend, RedisBackend


class DatabaseBackendTest(TestCase):

    def setUp(self):
        self.backend = DatabaseBackend()

    def test_add_delete_source(self):
        source_name = 'image.jpg'
        self.backend.add_source(source_name)
        self.assertTrue(Source.objects.filter(name=source_name).exists())
        self.backend.delete_source(source_name)
        self.assertFalse(Source.objects.filter(name=source_name).exists())

    def test_get_source(self):
        source_name = 'image.jpg'

        self.backend.add_source(source_name)
        self.assertEqual(
            self.backend.get_source(source_name),
            Source.objects.get(name=source_name)
        )

    def test_add_delete_thumbnail(self):
        source_name = 'image.jpg'
        self.backend.add_source(source_name)
        self.backend.add_thumbnail(source_name, 'small', 'image_small.jpg')

        source = self.backend.get_source(name=source_name)
        self.assertTrue(ThumbnailMeta.objects.filter(source=source, size='small').exists())
        self.backend.delete_thumbnail(source_name, 'small')
        self.assertFalse(ThumbnailMeta.objects.filter(source=source, size='small').exists())

    def test_get_thumbnail(self):
        source_name = 'image.jpg'
        self.backend.add_source(source_name)
        self.backend.add_thumbnail(source_name, 'small', 'image_small.jpg')
        self.assertEqual(self.backend.get_thumbnail(source_name, 'small'), ImageMeta(source_name, 'image_small.jpg', 'small'))
        self.backend.add_thumbnail(source_name, 'large', 'image_large.jpg')
        self.backend.add_thumbnail(source_name, 'source_with_format', 'image_source_with_format.webp')

        self.assertEqual(
            self.backend.get_thumbnails(source_name),
            [
                ImageMeta(source_name, 'image_small.jpg', 'small'),
                ImageMeta(source_name, 'image_large.jpg', 'large'),
                ImageMeta(source_name, 'image_source_with_format.webp', 'source_with_format')
            ]
        )

    def test_flush_thumbnails(self):
        source_name = 'image.jpg'
        self.backend.add_source(source_name)
        self.backend.add_thumbnail(source_name, 'small', 'image_small.jpg')
        self.backend.add_thumbnail(source_name, 'default', 'image_default.jpg')

        self.backend.flush_thumbnails(source_name)
        self.assertFalse(ThumbnailMeta.objects.filter(source__name=source_name).exists())


class RedisBackendTest(TestCase):

    def setUp(self):
        self.backend = RedisBackend()
        self.redis = StrictRedis()

    def test_get_source_key(self):
        self.assertEqual(self.backend.get_source_key('a.jpg'), 'djthumbs-test:sources:a.jpg')

    def test_get_thumbnail_key(self):
        self.assertEqual(self.backend.get_thumbnail_key('a.jpg'), 'djthumbs-test:thumbnails:a.jpg')

    def test_add_delete_source(self):
        source_name = 'test-thumbnail.jpg'
        source_key = self.backend.get_source_key(source_name)

        self.backend.add_source(source_name)
        self.assertTrue(self.redis.hexists(source_key, source_name))
        self.backend.delete_source(source_name)
        self.assertFalse(self.redis.hexists(source_key, source_name))

    def test_get_source(self):
        source_name = 'test-thumbnail.jpg'
        source_key = self.backend.get_source_key(source_name)

        self.redis.hset(source_key, source_name, source_name)
        self.assertEqual(self.backend.get_source(source_name), source_name)

        # Delete Source
        self.redis.hdel(source_key, source_name)

    def test_add_delete_thumbnail(self):
        source_name = 'test-thumbnail.jpg'
        size = 'small'
        thumbnail_key = self.backend.get_thumbnail_key(source_name)

        self.backend.add_source(source_name)
        self.backend.add_thumbnail(source_name, size, 'test-thumbnail_small.jpg')
        self.assertTrue(self.redis.hexists(thumbnail_key, size))

        self.backend.delete_thumbnail(source_name, size)
        self.assertFalse(self.redis.hexists(thumbnail_key, size))

        # Delete Source
        self.redis.hdel(self.backend.get_source_key(source_name), source_name)

    def test_get_thumbnail(self):
        source_name = 'test-thumbnail.jpg'

        self.backend.add_source(source_name)
        self.backend.add_thumbnail(source_name, 'small', 'test-thumbnail_small.jpg')
        self.assertEqual(self.backend.get_thumbnail(source_name, 'small'), ImageMeta(source_name, 'test-thumbnail_small.jpg', 'small'))
        self.backend.add_thumbnail(source_name, 'large', 'test-thumbnail_large.jpg')

        expected = ['test-thumbnail_large.jpg', 'test-thumbnail_small.jpg']
        result = [image_meta.name for image_meta in self.backend.get_thumbnails(source_name)]
        # sort is replacing the variable in place, not returning new value, it will always return None
        result.sort()
        expected.sort()
        self.assertEqual(result, expected)

        # Delete Source & Thumbnails
        thumbnail_key = self.backend.get_thumbnail_key(source_name)
        self.redis.hdel(self.backend.get_source_key(source_name), source_name)
        self.redis.hdel(thumbnail_key, 'small')
        self.redis.hdel(thumbnail_key, 'large')

    def test_flush_thumbnails(self):
        source_name = 'test-thumbnail.jpg'
        thumbnail_key = self.backend.get_thumbnail_key(source_name)

        self.backend.add_source(source_name)
        self.backend.add_thumbnail(source_name, "small", 'test-thumbnail_small.jpg')
        self.backend.add_thumbnail(source_name, "default", 'test-thumbnail_default.jpg')

        self.backend.flush_thumbnails(source_name)
        self.assertFalse(self.redis.exists(thumbnail_key))



import sys

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3


if PY2:
    string_types = (str, unicode)
    text_type = unicode

    def as_text(v):
        return v

else:
    string_types = (str,)
    text_type = str

    def as_text(v):
        if v is None:
            return None
        elif isinstance(v, bytes):
            return v.decode('utf-8')
        elif isinstance(v, str):
            return v
        else:
            raise ValueError('Unknown type %r' % type(v))



import os

from django.core.files import File
from django.test import TestCase

from thumbnails import conf
from thumbnails.backends.metadata import RedisBackend
from thumbnails.metadata import get_path
from thumbnails.models import ThumbnailMeta

from .models import TestModel


class FilesTest(TestCase):

    def setUp(self):
        self.source_name = "tests.png"
        self.size = "small"

        self.instance = TestModel.objects.create()
        with open('thumbnails/tests/tests.png', 'rb') as image_file:
            self.instance.avatar = File(image_file)
            self.instance.save()
        self.avatar_folder = \
            os.path.join(self.instance.avatar.storage.temporary_location, conf.BASE_DIR, 'avatars')

        self.instance.avatar.thumbnails.small

        self.basename = os.path.basename(self.instance.avatar.path)
        self.filename, self.ext = os.path.splitext(self.basename)
        self.small_metadata_name = os.path.join('thumbs', 'avatars', self.filename + "_" + self.size + self.ext)

    def tearDown(self):
        self.instance.avatar.storage.delete_temporary_storage()
        super(FilesTest, self).tearDown()

    def test_get_file_path(self):
        self.assertEqual(self.small_metadata_name, get_path(self.instance.avatar.name, self.size))

    def test_delete_without_thumbnails(self):
        # delete without thumbnails
        avatar_path = self.instance.avatar.path
        # ensure file still exists
        self.assertTrue(os.path.exists(avatar_path))

        self.instance.avatar.delete(with_thumbnails=False)

        # image file is deleted
        self.assertFalse(os.path.exists(avatar_path))
        # thumbnails and their metadata are not deleted
        self.assertEqual(len(os.listdir(self.avatar_folder)), 1)
        self.assertTrue(ThumbnailMeta.objects.filter(name=self.small_metadata_name).exists())

    def test_delete_with_thumbnails(self):
        avatar_path = self.instance.avatar.path
        # ensure file still exists
        self.assertTrue(os.path.exists(avatar_path))
        self.assertTrue(ThumbnailMeta.objects.filter(name=self.small_metadata_name).exists())

        self.instance.avatar.delete(with_thumbnails=True)

        # image file is deleted
        self.assertFalse(os.path.exists(avatar_path))

        # thumbnails and their metadata are also deleted
        self.assertEqual(len(os.listdir(self.avatar_folder)), 0)
        self.assertFalse(ThumbnailMeta.objects.filter(source__name=self.instance.avatar.name).exists())

    def test_delete_all(self):
        thumbnails = self.instance.avatar.thumbnails

        thumbnails.delete_all()
        # thumbnails and their metadata are deleted
        self.assertEqual(len(os.listdir(self.avatar_folder)), 0)
        self.assertFalse(ThumbnailMeta.objects.filter(source__name=self.instance.avatar.name).exists())


class RedisFilesTest(TestCase):
    def setUp(self):
        self.source_name = "tests.png"
        self.size = "small"

        self.instance = TestModel.objects.create()
        with open('thumbnails/tests/tests.png', 'rb') as image_file:
            self.instance.avatar = File(image_file)
            self.instance.save()
        self.backend = RedisBackend()
        self.instance.avatar.thumbnails.metadata_backend = self.backend
        self.avatar_folder = \
            os.path.join(self.instance.avatar.storage.temporary_location, conf.BASE_DIR, 'avatars')

        self.instance.avatar.thumbnails.small

        self.basename = os.path.basename(self.instance.avatar.path)
        self.filename, self.ext = os.path.splitext(self.basename)

    def tearDown(self):
        key = self.backend.get_thumbnail_key(self.instance.avatar.name)
        self.backend.redis.delete(key)
        self.instance.avatar.storage.delete_temporary_storage()

        super(RedisFilesTest, self).tearDown()

    def test_delete_with_thumbnails(self):
        avatar_path = self.instance.avatar.path
        # ensure file still exists
        self.assertTrue(os.path.exists(avatar_path))

        key = self.backend.get_thumbnail_key(self.instance.avatar.name)
        self.assertTrue(self.backend.redis.exists(key))

        self.instance.avatar.delete(with_thumbnails=True)

        # image file is deleted
        self.assertFalse(os.path.exists(avatar_path))

        # thumbnails and their metadata are also deleted
        self.assertEqual(len(os.listdir(self.avatar_folder)), 0)
        self.assertFalse(self.backend.redis.exists(key))

    def test_delete_all(self):
        thumbnails = self.instance.avatar.thumbnails
        key = self.backend.get_thumbnail_key(self.instance.avatar.name)
        self.assertTrue(self.backend.redis.exists(key))

        thumbnails.delete_all()
        # thumbnails and their metadata are deleted
        self.assertEqual(len(os.listdir(self.avatar_folder)), 0)
        self.assertFalse(self.backend.redis.exists(key))



import io

from django.core.files.base import ContentFile

from da_vinci import images


def resize(image, **kwargs):
    """
    Resized an image based on given width and height
    """
    image.resize(**kwargs)
    return image


def rotate(image, **kwargs):
    """
    Rotates an argument by X degrees.
    """
    image.rotate(**kwargs)
    return image


def flip(image, **kwargs):
    """
    Flips an image. Expects "horizontal" or "vertical" as argument.
    """
    image.flip(**kwargs)
    return image


def crop(image, **kwargs):
    """
    Crops an image based on given width or height
    """
    image.crop(**kwargs)
    return image


def set_quality(image, **kwargs):
    """
    Sets JPG images' quality.
    """
    image.quality = kwargs['quality']
    return image


def add_watermark(image, **kwargs):
    watermark_path = kwargs["watermark_path"]
    watermark_image = images.from_file(watermark_path)
    pil_image = image.get_pil_image()
    watermark_pil_image = watermark_image.get_pil_image()

    if watermark_pil_image.size != pil_image.size:
        # TODO: parse watermark dynamically based on ratio
        raise ValueError("Watermark image should have the same dimension as image")

    if watermark_image.format != "PNG" or watermark_pil_image.mode != "RGBA":
        raise ValueError("Watermark must be PNG and containts alpha")

    pil_image.paste(watermark_pil_image, (0, 0), watermark_pil_image)
    image.set_pil_image(pil_image)

    return image


def process(file, size):
    """
    Process an image through its defined processors
    params :file: filename or file-like object
    params :size: string for size defined in settings
    return a ContentFile
    """
    from . import conf
    # open image in piccaso
    raw_image = images.from_file(file)

    # run through all processors, if defined
    size_dict = conf.SIZES[size]

    # Format image if FORMAT defined in settings
    if 'FORMAT' in size_dict:
        raw_image.format = size_dict['FORMAT']

    for processor in size_dict['PROCESSORS']:
        raw_image = processor['processor'](raw_image, **processor['kwargs'])

    # write to Content File
    image_io = io.BytesIO()
    raw_image.save(file=image_io)
    image_file = ContentFile(image_io.getvalue())

    return image_file



import shortuuid
import os

from django.db.models import ImageField as DjangoImageField

from thumbnails import compat

from .backends import metadata, storage
from .backends.metadata import ImageMeta
from .files import ThumbnailedImageFile
from .images import Thumbnail, save
from . import processors, post_processors, conf


class ImageField(DjangoImageField):
    attr_class = ThumbnailedImageFile

    def __init__(self, *args, **kwargs):
        self.resize_source_to = kwargs.pop('resize_source_to', None)
        self.pregenerated_sizes = kwargs.pop('pregenerated_sizes', [])
        self.storage_backend = kwargs.get('storage') or storage.get_backend()
        self.metadata_backend = metadata.get_backend()
        kwargs['storage'] = self.storage_backend
        super(ImageField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(ImageField, self).deconstruct()
        del kwargs['storage']
        return name, path, args, kwargs

    def __unicode__(self):
        return self.attname

    def pre_save(self, model_instance, add):
        """
        Process the source image through the defined processors.
        """
        file = getattr(model_instance, self.attname)

        if file and not file._committed:
            image_file = file
            original_filename = file.name
            file_type = os.path.splitext(original_filename)[1]

            if self.resize_source_to:
                file.seek(0)
                image_file = processors.process(file, self.resize_source_to)
                image_file = post_processors.process(image_file, self.resize_source_to)

                if 'FORMAT' in conf.SIZES[self.resize_source_to]:
                    file_type = ".{}".format(conf.SIZES[self.resize_source_to]['FORMAT'])

            filename = str(shortuuid.uuid()) + file_type
            file.save(filename, image_file, save=False)

            for size in self.pregenerated_sizes:
                if size == self.resize_source_to:
                    # no need to process file if it is in resize_source_to,
                    # since it had been processed right before this
                    save(file.name, size, self.metadata_backend,
                         self.storage_backend, image_file)
                    continue
                file.thumbnails.create(size)

        return file

    def south_field_triple(self):
        """
        Return a suitable description of this field for South.
        Taken from smiley chris' easy_thumbnails
        """
        from south.modelsinspector import introspector
        field_class = 'django.db.models.fields.files.ImageField'
        args, kwargs = introspector(self)
        return (field_class, args, kwargs)


def fetch_thumbnails(images, sizes=None):
    """
    Regenerate EXISTING thumbnails, so we don't need to call redis when using
    thumbnails.get() or thumbnails.all(). Currently only support redis backend.
    NotImeplementedError will be raised, if backend is not supported
    """
    # NOTE: This is just working for redis based backend and same backend
    # different backend among thumbnails may results in bugs
    if not images:
        return

    backend = images[0].thumbnails.metadata_backend
    try:
        pipeline = backend.redis.pipeline()
    except AttributeError:
        raise NotImplementedError('Only Redis metadata backend is implemented')

    for image in images:
        thumbnails = image.thumbnails
        key = thumbnails.metadata_backend.get_thumbnail_key(thumbnails.source_image.name)
        if sizes:
            pipeline.hmget(key, sizes)
        else:
            pipeline.hgetall(key)

    # if sizes is provided results will be list of lists, else it will be list of dicts
    results = pipeline.execute()
    for image, data in zip(images, results):
        thumbnails = image.thumbnails
        source_name = thumbnails.source_image.name
        thumbnails._thumbnails = {}

        if sizes:
            # data shold be list, thus group it with its size beforehand
            items = zip(sizes, data)
        else:
            # data should be dict
            items = data.items()

        for size, name in items:
            if not name:
                continue
            image_meta = ImageMeta(source_name, name, size)
            thumbnails._thumbnails[compat.as_text(size)] = Thumbnail(image_meta, thumbnails.storage)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Source',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(unique=True, max_length=255)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='ThumbnailMeta',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('size', models.CharField(max_length=64)),
                ('name', models.CharField(unique=True, max_length=255)),
                ('source', models.ForeignKey(on_delete=models.deletion.CASCADE, related_name='thumbnails', to='thumbnails.Source')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='thumbnailmeta',
            unique_together=set([('source', 'size')]),
        ),
    ]



import os
import tempfile
import shutil

from django.core.files.storage import FileSystemStorage

"""
Temporary Storage class for test. Copied from Smiley Chris' Easy Thumbnails test package
https://github.com/SmileyChris/easy-thumbnails/blob/master/easy_thumbnails/test.py
"""


class TemporaryStorage(FileSystemStorage):
    """
    A storage class useful for tests that uses a temporary location to store
    all files and provides a method to remove this location when it is finished
    with.
    """

    def __init__(self, location=None, *args, **kwargs):
        """
        Create the temporary location.
        """
        if location is None:
            location = tempfile.mkdtemp()
            self.temporary_location = location

        super(TemporaryStorage, self).__init__(location=location, *args,
                                               **kwargs)

    def delete_temporary_storage(self):
        """
        Delete the temporary directory created during initialisation.
        This storage class should not be used again after this method is
        called.
        """
        temporary_location = getattr(self, 'temporary_location', None)
        if temporary_location:
            shutil.rmtree(temporary_location)


class TemporaryStorage2(TemporaryStorage):
    pass



from django.test import TestCase
from django.core.files import File

from thumbnails.post_processors import optimize


class PostProcessorsTest(TestCase):

    def test_optimize(self):
        with open('thumbnails/tests/tests.png', 'rb') as image_file:
            original_file = File(image_file)
            optimized_file = optimize(
                original_file.file,
                png_command="optipng -force -o7 '%(filename)s' "
            )
            self.assertTrue(original_file.size > optimized_file.size)



from django.apps import apps
from django.core.management.base import BaseCommand

from thumbnails import images
from thumbnails.backends import metadata, storage


class Command(BaseCommand):
    """
    Delete all thumbnails from a field. For example:
    python manage.py delete_thumbnails --model='accounts.Profile' --field='picture' --size='small'
    """

    def add_arguments(self, parser):
        parser.add_argument('--model', dest='path_to_model',
                            help='Dotted path to model (e.g "polls.Poll" )')
        parser.add_argument('--field', dest='field_name',
                            help='Thumbnail field name (e.g "picture"')
        parser.add_argument('--size', dest='size',
                            help='Thumbnail size to delete (e.g "small"')

    def handle(self, path_to_model, field_name, size, *args, **kwargs):
        if not path_to_model:
            raise ValueError('--model argument is required')
        if not field_name:
            raise ValueError('--field argument is required')
        if not size:
            raise ValueError('--size argument is required')

        app_label, model_name = path_to_model.rsplit('.', 1)
        model = apps.get_model(app_label, model_name)

        # Get model instances which has non empty fields
        exclude_args = {
            '%s__isnull' % field_name: True,
            field_name: ''
        }
        instances = model.objects.exclude(**exclude_args)

        metadata_backend = metadata.get_backend()

        for instance in instances:
            field = getattr(instance, field_name)
            images.delete(field.name, size, metadata_backend, field.thumbnails.storage)



# -*- coding: utf-8 -*-
from setuptools import setup

setup(
    name='django-thumbnails',
    version='0.7.0',
    author='Selwin Ong',
    author_email='selwin.ong@gmail.com',
    packages=['thumbnails'],
    url='https://github.com/ui/django-thumbnails',
    license='MIT',
    description='A simple Django app to manage image/photo thumbnails. Supports remote/cloud storage systems like Amazon S3.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    zip_safe=False,
    include_package_data=True,
    package_data={'': ['README.md']},
    install_requires=['django>=2.0', 'da-vinci', 'shortuuid', 'pillow'],
    extras_require = {
        'redis':  ['redis']
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



from django.db import models

from thumbnails.fields import ImageField
from . import storage


class TestModel(models.Model):
    avatar = ImageField(upload_to='avatars', resize_source_to='source')
    profile_picture = ImageField(upload_to='avatars', blank=True, null=True, resize_source_to='source')
    card_identity_picture = ImageField(upload_to='identity_card', blank=True, null=True,
                                       resize_source_to='source_with_format',
                                       storage=storage.TemporaryStorage2())


class TestPregeneratedSizesModel(models.Model):
    logo = ImageField(upload_to='logos', blank=True, null=True, pregenerated_sizes=['small', 'large'])
    photo = ImageField(upload_to='photos', blank=True, null=True, resize_source_to='source_with_format',
                       pregenerated_sizes=['source_with_format', 'default', 'large'])



from thumbnails import conf
from thumbnails.utils import import_attribute


def get_backend():
    if not conf.STORAGE.get('BACKEND'):
        raise ValueError('BACKEND for STORAGE must be defined')
    backend = import_attribute(conf.STORAGE['BACKEND'])
    return backend()



import imghdr
import os
from subprocess import call
import tempfile
import shortuuid

from django.core.files import File


def get_or_create_temp_dir():
    temp_dir = os.path.join(tempfile.gettempdir(), 'thumbnails')
    if not os.path.exists(temp_dir):
        os.mkdir(temp_dir)
    return temp_dir


def process(thumbnail_file, size, **kwargs):
    """
    Post processors are functions that receive file objects,
    performs necessary operations and return the results as file objects.
    """
    from . import conf

    size_dict = conf.SIZES[size]
    for processor in size_dict['POST_PROCESSORS']:
        processor['processor'](thumbnail_file, **processor['kwargs'])

    return thumbnail_file


def optimize(thumbnail_file, jpg_command=None, png_command=None,
             gif_command=None):
    """
    A post processing function to optimize file size. Accepts commands
    to optimize JPG, PNG and GIF images as arguments. Example:

    THUMBNAILS = {
        # Other options...
        'POST_PROCESSORS': [
            {
                'processor': 'thumbnails.post_processors.optimize',
                'png_command': 'optipng -force -o3 "%(filename)s"',
                'jpg_command': 'jpegoptim -f --strip-all "%(filename)s"',
            },
        ],
    }

    Note: using output redirection in commands may cause unpredictable results.
    For example 'optipng -force -o3 "%(filename)s" &> /dev/null' may cause
    optimize command to fail on some systems.
    """
    temp_dir = get_or_create_temp_dir()
    thumbnail_filename = os.path.join(temp_dir, "%s" % shortuuid.uuid())

    f = open(thumbnail_filename, 'wb')
    f.write(thumbnail_file.read())
    f.close()

    # Detect filetype
    filetype = imghdr.what(thumbnail_filename)

    # Construct command to optimize image based on filetype
    command = None
    if filetype == "jpg" or filetype == "jpeg":
        command = jpg_command
    elif filetype == "png":
        command = png_command
    elif filetype == "gif":
        command = gif_command

    # Run Command
    if command:
        command = command % {'filename': thumbnail_filename}
        call(command, shell=True)

    optimized_file = File(open(thumbnail_filename, 'rb'))
    os.remove(thumbnail_filename)

    return optimized_file



from .backends import metadata


def get_path(source_name, size=None):
    if size is None:
        instance = metadata.get_backend().get_source(source_name)
    else:
        instance = metadata.get_backend().get_thumbnail(source_name, size)

    if instance is None:
        return instance
    else:
        return instance.name



from django.conf import settings

from .utils import parse_processors

THUMBNAILS = getattr(settings, 'THUMBNAILS', {})


default_metadata = {
    'BACKEND': 'thumbnails.backends.metadata.DatabaseBackend'
}

default_storage = {
    'BACKEND': 'django.core.files.storage.FileSystemStorage'
}

METADATA = THUMBNAILS.get('METADATA', default_metadata)
STORAGE = THUMBNAILS.get('STORAGE', default_storage)
SIZES = THUMBNAILS.get('SIZES', {})
BASE_DIR = THUMBNAILS.get('BASE_DIR', 'thumbnails')


# import the processors as a functions and replace the import string
for size in SIZES:
    if SIZES[size].get('PROCESSORS') is not None:
        SIZES[size]['PROCESSORS'] = parse_processors(SIZES[size]['PROCESSORS'])
    else:
        SIZES[size]['PROCESSORS'] = []

    if SIZES[size].get('POST_PROCESSORS') is not None:
        SIZES[size]['POST_PROCESSORS'] = parse_processors(SIZES[size]['POST_PROCESSORS'])
    else:
        SIZES[size]['POST_PROCESSORS'] = []
